"""
Types for encoding and decoding data.
"""

from typing import (
    NewType,
    Union,
)

HexStr = NewType("HexStr", str)
"""
A string that represents a hex encoded value.
"""
Primitives = Union[bytes, int, bool]
"""
A type that represents bytes, int or bool primitive types.
"""



"""
Types used for BLS Signatures.
"""

from typing import (
    NewType,
)

BLSPubkey = NewType("BLSPubkey", bytes)
"""
A BLS public key that is 48 bytes in length.
"""
BLSPrivateKey = NewType("BLSPrivateKey", int)
"""
A BLS private key integer value.
"""
BLSSignature = NewType("BLSSignature", bytes)
"""
A BLS signature that is 96 bytes in length.
"""



"""
Exception types raised in web3's libraries.
"""


class ValidationError(Exception):
    """
    Raised when something does not pass a validation check.
    """


class MismatchedABI(ValidationError):
    """
    Raised when an ABI does not match with supplied parameters, or when an
    attempt is made to access a function/event that does not exist in the ABI.
    """



"""
Types for the Discovery Protocol.
"""

from typing import (
    NewType,
)

NodeID = NewType("NodeID", bytes)
r"""
A 32-byte identifier for a node in the Discovery DHT.

.. doctest::

    >>> from eth_typing import NodeID
    >>> NodeID(b'\x01' * 32)
    b'\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01'
"""



"""
Types for Contract ABIs and related components.
"""

from collections.abc import (
    Sequence,
)
from typing import (
    Any,
    Literal,
    TypedDict,
    Union,
)

from typing_extensions import (
    NotRequired,
)

from eth_typing.encoding import (
    HexStr,
)

TypeStr = str
"""String representation of a data type."""
Decodable = Union[bytes, bytearray]
"""Binary data to be decoded."""


class ABIComponent(TypedDict):
    """
    TypedDict representing an `ABIElement` component.
    """

    type: str
    """Type of the component."""
    name: NotRequired[str]
    """Name of the component."""
    components: NotRequired[Sequence["ABIComponent"]]
    """List of nested `ABI` components for ABI types."""


class ABIComponentIndexed(ABIComponent):
    """
    TypedDict representing an indexed `ABIElement` component.
    """

    indexed: bool
    """If True, component can be used as a topic filter."""


class ABIEvent(TypedDict):
    """
    TypedDict to represent the `ABI` for an event.
    """

    name: str
    """Event name identifier."""
    type: Literal["event"]
    """Event ABI type."""
    anonymous: NotRequired[bool]
    """If True, event is anonymous. Cannot filter the event by name."""
    inputs: NotRequired[Sequence["ABIComponentIndexed"]]
    """Input components for the event."""


class ABIFunctionType(TypedDict, total=False):
    """
    TypedDict representing the `ABI` for all function types.

    This is the base type for functions.
    Please use ABIFunction, ABIConstructor, ABIFallback or ABIReceive instead.
    """

    stateMutability: Literal["pure", "view", "nonpayable", "payable"]
    """State mutability of the constructor."""
    payable: bool
    """
    Contract is payable to receive ether on deployment.
    Deprecated in favor of stateMutability payable and nonpayable.
    """
    constant: bool
    """
    Function is constant and does not change state.
    Deprecated in favor of stateMutability pure and view.
    """


class ABIFunction(ABIFunctionType):
    """
    TypedDict representing the `ABI` for a function.
    """

    type: Literal["function"]
    """Type of the function."""
    name: str
    """Name of the function."""
    inputs: NotRequired[Sequence["ABIComponent"]]
    """Function input components."""
    outputs: NotRequired[Sequence["ABIComponent"]]
    """Function output components."""


class ABIConstructor(ABIFunctionType):
    """
    TypedDict representing the `ABI` for a constructor function.
    """

    type: Literal["constructor"]
    """Type of the constructor function."""
    inputs: NotRequired[Sequence["ABIComponent"]]
    """Function input components."""


class ABIFallback(ABIFunctionType):
    """
    TypedDict representing the `ABI` for a fallback function.
    """

    type: Literal["fallback"]
    """Type of the fallback function."""


class ABIReceive(ABIFunctionType):
    """
    TypedDict representing the `ABI` for a receive function.
    """

    type: Literal["receive"]
    """Type of the receive function."""


class ABIError(TypedDict):
    """
    TypedDict representing the `ABI` for an error.
    """

    type: Literal["error"]
    """Type of the error."""
    name: str
    """Name of the error."""
    inputs: NotRequired[Sequence["ABIComponent"]]
    """Error input components."""


ABICallable = Union[ABIFunction, ABIConstructor, ABIFallback, ABIReceive]
"""
A `Union` type consisting of `ABIFunction`, `ABIConstructor`, `ABIFallback` and
`ABIReceive` types.
"""

ABIElement = Union[ABICallable, ABIEvent, ABIError]
"""A `Union` type consisting of `ABICallable`, `ABIEvent`, and `ABIError` types."""


class ABIElementInfo(TypedDict):
    """
    TypedDict to represent properties of an `ABIElement`, including the abi,
    selector and arguments.
    """

    abi: ABIElement
    """ABI for any `ABIElement` type."""
    selector: HexStr
    """Solidity `ABIElement` selector sighash."""
    arguments: tuple[Any, ...]
    """`ABIElement` input components."""


ABI = Sequence[ABIElement]
"""
List of components representing function and event interfaces
(elements of an ABI).
"""



"""
Type definitions for the Ethereum Virtual Machine (EVM).
"""

from typing import (
    Literal,
    NewType,
    TypeVar,
    Union,
)

from .encoding import (
    HexStr,
)

Hash32 = NewType("Hash32", bytes)
"""
A 32-byte hash value.
"""
BlockNumber = NewType("BlockNumber", int)
"""
Any integer that represents a valid block number on a chain.
"""
BlockParams = Literal["latest", "earliest", "pending", "safe", "finalized"]
"""
A type which specifies the block reference parameter.

- ``"latest"``: The latest block.
- ``"earliest"``: The earliest block.
- ``"pending"``: The pending block.
- ``"safe"``: The safe block.
- ``"finalized"``: The finalized block.
"""
BlockIdentifier = Union[BlockParams, BlockNumber, Hash32, HexStr, int]
"""
A type that represents a block identifier value.

- ``BlockParams``: A block reference parameter.
- ``BlockNumber``: A block number integer value.
- ``Hash32``: A 32-byte hash value.
- ``HexStr``: A string that represents a hex value.
- ``int``: An integer value.
"""
Address = NewType("Address", bytes)
"""
A type that contains a 32-byte canonical address.
"""
HexAddress = NewType("HexAddress", HexStr)
"""
A type that contains a hex encoded address. This is a 32-byte hex string with a prefix
of "0x".
"""
ChecksumAddress = NewType("ChecksumAddress", HexAddress)
"""
A type that contains a eth_typing.evm.HexAddress that is formatted according to
`ERC55 <https://github.com/ethereum/EIPs/issues/55>`_. This is a 40 character hex
string with a prefix of "0x" and mixed case letters.
"""
AnyAddress = TypeVar("AnyAddress", Address, HexAddress, ChecksumAddress)
"""
A type that represents any type of address.
"""
